/*
 * Copyright (C) 2011-2012 by Matthias Ringwald
 * Copyright (C) 2013 by Kunihiro Shibuya
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 4. This software may not be used in a commercial product
 *    without an explicit license granted by the copyright holder. 
 *
 * THIS SOFTWARE IS PROVIDED BY MATTHIAS RINGWALD AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL MATTHIAS
 * RINGWALD OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

//*****************************************************************************
//
// att device demo
//
//*****************************************************************************

// TODO: seperate BR/EDR from LE ACL buffers
// TODO: move LE init into HCI
// ..

// NOTE: Supports only a single connection

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "global.h"

#include <btstack/run_loop.h>

#include "btstack_memory.h"
#include "hci.h"
#include "hci_dump.h"
#include "l2cap.h"
#include "btstack/debug.h"

#include "att.h"

hci_transport_t * hci_transport_picusb_instance();

//Advertise data
const uint8_t adv_data[31]=
	"\x02\x01\x05"
	"\x06\x09SBBLE"
	"\x03\x02\xf0\xff"
	;

static att_connection_t att_connection;
static uint16_t         att_response_handle = 0;
static uint16_t         att_response_size   = 0;
static uint8_t          att_response_buffer[28];

uint8_t	connection_status=0;

static void initconfig(void);	//sibu

void hexdump2(void *data, int size){
    int i;
    for (i=0; i<size;i++){
        log_info("%02X ", ((uint8_t *)data)[i]);
    }
    log_info("\n");
}

static void att_try_respond(void){
    if (!att_response_size) return;
    if (!att_response_handle) return;
    if (!hci_can_send_packet_now(HCI_ACL_DATA_PACKET)) return;
    
    // update state before sending packet
    uint16_t size = att_response_size;
    att_response_size = 0;
    l2cap_send_connectionless(att_response_handle, L2CAP_CID_ATTRIBUTE_PROTOCOL, att_response_buffer, size);
}


static void att_packet_handler(uint8_t packet_type, uint16_t handle, uint8_t *packet, uint16_t size){
    if (packet_type != ATT_DATA_PACKET) return;
    
    att_response_handle = handle;
    att_response_size = att_handle_request(&att_connection, packet, size, att_response_buffer);
    att_try_respond();
}


// enable LE, setup ADV data
static void packet_handler (void * connection, uint8_t packet_type, uint16_t channel, uint8_t *packet, uint16_t size){
    static bd_addr_t addr;
    switch (packet_type) {
            
		case HCI_EVENT_PACKET:
			switch (packet[0]) {
				
                case BTSTACK_EVENT_STATE:
					// bt stack activated, get started - set local name
					if (packet[2] == HCI_STATE_WORKING) {
					   log_info("Working!\n");
						hci_send_cmd(&hci_read_local_supported_features);
					}
					break;
                
                case DAEMON_EVENT_HCI_PACKET_SENT:
                    att_try_respond();
                    break;
                    
                case HCI_EVENT_LE_META:
                    switch (packet[2]) {
                        case HCI_SUBEVENT_LE_CONNECTION_COMPLETE:
                            // reset connection MTU
                            att_connection.mtu = 23;
                            break;
                        default:
                            break;
                    }
                    break;

                case BTSTACK_EVENT_NR_CONNECTIONS_CHANGED:
				    if (packet[2]) {
						connection_status=1;
                        log_info("CONNECTED\n");
                    } else {
						connection_status=0;
						initconfig();				//sibu
				        log_info("NOT CONNECTED\n");
                    }
                    break;
                    
                case HCI_EVENT_DISCONNECTION_COMPLETE:
                    att_response_handle =0;
                    att_response_size = 0;
                    
                    // restart advertising
                    hci_send_cmd(&hci_le_set_advertise_enable, 1);
                    break;
                    
				case HCI_EVENT_COMMAND_COMPLETE:
					if (COMMAND_COMPLETE_EVENT(packet, hci_read_bd_addr)){
    					bt_flip_addr(addr, &packet[6]);
					    log_info("BD ADDR: %s\n", bd_addr_to_str(addr));
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_read_local_supported_features)){
					    log_info("Local supported features: %04lX%04lX\n", READ_BT_32(packet, 10), READ_BT_32(packet, 6));
                        hci_send_cmd(&hci_set_event_mask, 0xffffffff, 0x20001fff);
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_set_event_mask)){
                        hci_send_cmd(&hci_write_le_host_supported, 1, 1);
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_write_le_host_supported)){
				        hci_send_cmd(&hci_le_set_event_mask, 0xffffffff, 0xffffffff);
                    	break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_event_mask)){
				        hci_send_cmd(&hci_le_read_buffer_size);
                    	break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_read_buffer_size)){
					    log_info("LE buffer size: %u, count %u\n", READ_BT_16(packet,6), packet[8]);
					   hci_send_cmd(&hci_le_read_supported_states);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_read_supported_states)){
					   //hci_send_cmd(&hci_le_set_advertising_parameters,  0x0400, 0x0800, 0, 0, 0, &addr, 0x07, 0);	//640-1280ms
					   hci_send_cmd(&hci_le_set_advertising_parameters,  0x00a0, 0x0140, 0, 0, 0, &addr, 0x07, 0);		//100-200ms
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_advertising_parameters)){
					   hci_send_cmd(&hci_le_set_advertising_data, sizeof(adv_data), adv_data);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_advertising_data)){
					   hci_send_cmd(&hci_le_set_scan_response_data, 10, adv_data);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_scan_response_data)){
					   hci_send_cmd(&hci_le_set_advertise_enable, 1);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_advertise_enable)){
                        hci_discoverable_control(1);
						log_info("startup_state=1\n");
						startup_state=1;
					    led1_off();
                        break;
					}
                    
			}
	}
}

// gatt profile
#include "profile.h"
#include "attribute.h"

#define	FLAG_NOTIFICATION			0x01

#define	GPIO_RB00					(16*1+ 0)
#define	GPIO_RB01					(16*1+ 1)
#define	GPIO_RB02					(16*1+ 2)
#define	GPIO_RB03					(16*1+ 3)
#define	GPIO_RB04					(16*1+ 4)
#define	GPIO_RB05					(16*1+ 5)
#define	GPIO_RB06					(16*1+ 6)
#define	GPIO_RB07					(16*1+ 7)
#define	GPIO_RB08					(16*1+ 8)
#define	GPIO_RB09					(16*1+ 9)
#define	GPIO_RB10					(16*1+10)
#define	GPIO_RB11					(16*1+11)
#define	GPIO_RB12					(16*1+12)
#define	GPIO_RB13					(16*1+13)
#define	GPIO_RB14					(16*1+14)
#define	GPIO_RB15					(16*1+15)
#define	GPIO_RC00					(16*2+ 0)
#define	GPIO_RC01					(16*2+ 1)
#define	GPIO_RC02					(16*2+ 2)
#define	GPIO_RC03					(16*2+ 3)
#define	GPIO_RC04					(16*2+ 4)
#define	GPIO_RC05					(16*2+ 5)
#define	GPIO_RC06					(16*2+ 6)
#define	GPIO_RC07					(16*2+ 7)
#define	GPIO_RC08					(16*2+ 8)
#define	GPIO_RC09					(16*2+ 9)
#define	GPIO_RC10					(16*2+10)
#define	GPIO_RC11					(16*2+11)
#define	GPIO_RC12					(16*2+12)
#define	GPIO_RC13					(16*2+13)
#define	GPIO_RC14					(16*2+14)
#define	GPIO_RC15					(16*2+15)

#define	GPIO_CN00					(16*0+ 0)
#define	GPIO_CN01					(16*0+ 1)
#define	GPIO_CN02					(16*0+ 2)
#define	GPIO_CN03					(16*0+ 3)
#define	GPIO_CN04					(16*0+ 4)
#define	GPIO_CN05					(16*0+ 5)
#define	GPIO_CN06					(16*0+ 6)
#define	GPIO_CN07					(16*0+ 7)
#define	GPIO_CN08					(16*0+ 8)
#define	GPIO_CN09					(16*0+ 9)
#define	GPIO_CN10					(16*0+10)
#define	GPIO_CN11					(16*0+11)
#define	GPIO_CN12					(16*0+12)
#define	GPIO_CN13					(16*0+13)
#define	GPIO_CN14					(16*0+14)
#define	GPIO_CN15					(16*0+15)
#define	GPIO_CN16					(16*1+ 0)
#define	GPIO_CN17					(16*1+ 1)
#define	GPIO_CN18					(16*1+ 2)
#define	GPIO_CN19					(16*1+ 3)
#define	GPIO_CN20					(16*1+ 4)
#define	GPIO_CN21					(16*1+ 5)
#define	GPIO_CN22					(16*1+ 6)
#define	GPIO_CN23					(16*1+ 7)
#define	GPIO_CN24					(16*1+ 8)
#define	GPIO_CN25					(16*1+ 9)
#define	GPIO_CN26					(16*1+10)
#define	GPIO_CN27					(16*1+11)
#define	GPIO_CN28					(16*1+12)
#define	GPIO_CN29					(16*1+13)
#define	GPIO_CN30					(16*1+14)
#define	GPIO_CN31					(16*1+15)

#define	GPIO_RP00					(16*0+ 0)
#define	GPIO_RP01					(16*0+ 1)
#define	GPIO_RP09					(16*0+ 9)
#define	GPIO_RP14					(16*0+14)
#define	GPIO_RP15					(16*0+15)
#define	GPIO_RP22					(16*0+22)
#define	GPIO_RP23					(16*0+23)
#define	GPIO_RP24					(16*0+24)
#define	GPIO_RP25					(16*0+25)

typedef struct{
	uint8_t		enable;
	uint8_t		invert;
	uint8_t		gpioport;
	uint8_t		cnport;
	uint8_t		rpno;
	uint8_t		usei2c;
} PORTMAP;

/*const*/ static PORTMAP portmap[]={
	{1,0,GPIO_RB14,GPIO_CN12,GPIO_RP14,0,},	//SBDBT LED1@ RB14,CN12
//	{1,0,GPIO_RB15,GPIO_CN11,GPIO_RP15,0,},	//SBDBT LED2  RB15,CN11
#if __DEBUG
//	{0,0,GPIO_RB00,GPIO_CN04,GPIO_RP00,0,},	//SBDBT pin4  RB0 ,CN4
//	{0,0,GPIO_RB01,GPIO_CN05,GPIO_RP01,0,},	//SBDBT pin5  RB1 ,CN5
#else
//	{1,0,GPIO_RB00,GPIO_CN04,GPIO_RP00,0,},	//SBDBT pin4  RB0 ,CN4
//	{1,0,GPIO_RB01,GPIO_CN05,GPIO_RP01,0,},	//SBDBT pin5  RB1 ,CN5
#endif
	{1,0,GPIO_RC09,GPIO_CN19,GPIO_RP25,0,},	//SBDBT pin6  RC9 ,CN19
	{1,0,GPIO_RC06,GPIO_CN18,GPIO_RP22,0,},	//SBDBT pin7  RC6 ,CN18
	{1,0,GPIO_RC07,GPIO_CN17,GPIO_RP23,1,},	//SBDBT pin8  RC7 ,CN17
	{1,0,GPIO_RC08,GPIO_CN20,GPIO_RP24,1,},	//SBDBT pin9  RC8 ,CN20
#if __DEBUG
	{0,0,GPIO_RC09,GPIO_CN21,GPIO_RP09,0,},	//SBDBT pin10 RC9 ,CN21
#else
	{1,0,GPIO_RB09,GPIO_CN21,GPIO_RP09,0,},	//SBDBT pin10 RC9 ,CN21
#endif
};

#define	PORTMAPS		6
#define	UART_MAPNO1		2
#define	UART_MAPNO2		3

static uint8_t	pio_setting=0;
static uint8_t	pio_pullup=0;
static uint8_t	pio_input=0;
static uint8_t	uart_config=0;
static uint16_t	uart_baudrate=UART_DEFAULT_BAUDRATE/240;
static uint8_t	uart_lasttx_size=0;
static uint8_t	uart_lasttx_str[20];
static uint32_t	pwm_period[8];
static uint32_t	pwm_duty[8];
static uint8_t	pwm_config;
uint8_t			uart1_polarity;
static uint8_t	i2c_config;
static uint8_t	i2c_recvbuf[20];
static uint8_t	i2c_recvlen;

//characteristic config
static uint8_t	pio_input_config=0;
static uint8_t	uart_rx_config=0;
static uint8_t	lowbat_notify_config=0;

#define	ADIN_VREF_VALUE		3300LU
#define	ADIN_FULLSCALE		1023LU

#if	1
#define	PWM_PRESCALER		T2_PS_1_256							// 16 MHz/256
#define	PWM_TIMER_PERIOD	1									//  1 count
#define	PWM_TIMER_RES		(1000000/(16000000/256))			// 16 us
#define	PWM_RATIO_US		(PWM_TIMER_RES*PWM_TIMER_PERIOD)	// 16 us
#else
#define	PWM_PRESCALER		T2_PS_1_64							// 16 MHz/64
#define	PWM_TIMER_PERIOD	5									//  5 count
#define	PWM_TIMER_RES		(1000000/(16000000/64))				//  4 us
#define	PWM_RATIO_US		(PWM_TIMER_RES*PWM_TIMER_PERIOD)	// 20 us
#endif

uint16_t	advalue2mv(uint16_t advalue)
{
	uint16_t	ret;

	ret=(uint16_t)((((uint32_t)advalue)*ADIN_VREF_VALUE)/ADIN_FULLSCALE);
	return ret;
}

void gpio_setup(uint8_t setup)
{
	int			port;
	unsigned	bitmask;
	int			mapno;
	int			output;

	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;
		}
		port=portmap[mapno].gpioport/16;
		bitmask=1<<(portmap[mapno].gpioport%16);
		output=setup&(1<<mapno);
	
		switch(port){
		case	0:	//PORTA
			if(output){
				TRISA&=~bitmask;
			}else{
				TRISA|=bitmask;
			}
			break;
		case	1:	//PORTB
			if(output){
				TRISB&=~bitmask;
			}else{
				TRISB|=bitmask;
			}
			break;
		case	2:	//PORTC
			if(output){
				TRISC&=~bitmask;
			}else{
				TRISC|=bitmask;
			}
			break;
		default:
			break;
		}
	}
	return;
}

void ppssetup(uint8_t rpno,uint8_t funcno)
{
	switch(rpno){
	case	GPIO_RP00:
		OUT_PIN_PPS_RP0=funcno;
		break;
	case	GPIO_RP01:
		OUT_PIN_PPS_RP1=funcno;
		break;
	case	GPIO_RP09:
		OUT_PIN_PPS_RP9=funcno;
		break;
	case	GPIO_RP14:
		OUT_PIN_PPS_RP14=funcno;
		break;
	case	GPIO_RP15:
		OUT_PIN_PPS_RP15=funcno;
		break;
	case	GPIO_RP22:
		OUT_PIN_PPS_RP22=funcno;
		break;
	case	GPIO_RP23:
		OUT_PIN_PPS_RP23=funcno;
		break;
	case	GPIO_RP24:
		OUT_PIN_PPS_RP24=funcno;
		break;
	case	GPIO_RP25:
		OUT_PIN_PPS_RP25=funcno;
		break;
	default:
		break;
	}
	return;
}

uint8_t gpio_input(void)
{
	uint8_t		data=0;
	int			port;
	unsigned	bitmask;
	int			mapno;

	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;					//ɂȂĂ
		}
		if(pio_setting&(1<<mapno)){
			continue;					//o͐ݒɂȂĂ
		}
		if(pwm_config&(1<<mapno)){
			continue;					//PWMo͂ɂȂĂ
		}
		if(i2c_config&&portmap[mapno].usei2c){
			continue;					//I2CŎgpĂ
		}
		port=portmap[mapno].gpioport/16;
		bitmask=1<<(portmap[mapno].gpioport%16);
	
		switch(port){
		case	0:	//PORTA
			if(PORTA&bitmask){
				data|=(1<<mapno);
			}
			break;
		case	1:	//PORTB
			if(PORTB&bitmask){
				data|=(1<<mapno);
			}
			break;
		case	2:	//PORTC
			if(PORTC&bitmask){
				data|=(1<<mapno);
			}
			break;
		default:
			break;
		}
		if(portmap[mapno].invert){
			data^=(1<<mapno);
		}
	}

	return data;
}

void gpio_output(uint8_t data)
{
	int			port;
	unsigned	bitmask;
	int			mapno;
	int			bitdata;

	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;
		}
		if(i2c_config&&portmap[mapno].usei2c){
			continue;					//I2CŎgpĂ
		}
		port=portmap[mapno].gpioport/16;
		bitmask=1<<(portmap[mapno].gpioport%16);
		bitdata=data&(1<<mapno);
		if(portmap[mapno].invert){
			bitdata=!bitdata;
		}
	
		switch(port){
		case	0:	//PORTA
			if(bitdata){
				LATA|=bitmask;
			}else{
				LATA&=~bitmask;
			}
			break;
		case	1:	//PORTB
			if(bitdata){
				LATB|=bitmask;
			}else{
				LATB&=~bitmask;
			}
			break;
		case	2:	//PORTC
			if(bitdata){
				LATC|=bitmask;
			}else{
				LATC&=~bitmask;
			}
			break;
		default:
			break;
		}
	}
	return;
}

void gpio_setpullup(uint8_t pullupsetting)
{
	int			port;
	unsigned	bitmask;
	int			mapno;
	int			pullup;

	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;
		}
		port=portmap[mapno].cnport/16;
		bitmask=1<<(portmap[mapno].cnport%16);
		pullup=pullupsetting&(1<<mapno);
	
		switch(port){
		case	0:	//CN1
			if(pullup){
				CNPU1|=bitmask;
			}else{
				CNPU1&=~bitmask;
			}
			break;
		case	1:	//CN2
			if(pullup){
				CNPU2|=bitmask;
			}else{
				CNPU2&=~bitmask;
			}
			break;
		default:
			break;
		}
	}
	return;
}

static void pwm_setup(uint8_t setup)
{
	int			port;
	unsigned	bitmask;
	int			mapno;
	int			output;

	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;
		}
		port=portmap[mapno].gpioport/16;
		bitmask=1<<(portmap[mapno].gpioport%16);
		output=setup&(1<<mapno);
	
		switch(port){
		case	0:	//PORTA
			if(output){
				TRISA&=~bitmask;
				ppssetup(portmap[mapno].rpno,OUT_FN_PPS_OC1+mapno);
			}else{
				if(!(pio_setting&(1<<mapno))){
					TRISA|=bitmask;
				}
				ppssetup(portmap[mapno].rpno,0x1f);
			}
			break;
		case	1:	//PORTB
			if(output){
				TRISB&=~bitmask;
				ppssetup(portmap[mapno].rpno,OUT_FN_PPS_OC1+mapno);
			}else{
				if(!(pio_setting&(1<<mapno))){
					TRISB|=bitmask;
				}
				ppssetup(portmap[mapno].rpno,0x1f);
			}
			break;
		case	2:	//PORTC
			if(output){
				TRISC&=~bitmask;
				ppssetup(portmap[mapno].rpno,OUT_FN_PPS_OC1+mapno);
			}else{
				if(!(pio_setting&(1<<mapno))){
					TRISC|=bitmask;
				}
				ppssetup(portmap[mapno].rpno,0x1f);
			}
			break;
		default:
			break;
		}
	}
	return;
}

#define	us2count(x)		((x)/PWM_RATIO_US)

static void set_pwm(int ch)
{
	uint32_t	period,duty;

	if(ch>=5){
		return;
	}
	period=us2count(pwm_period[ch]);
	duty=us2count(pwm_duty[ch]);

	if(period<2){
		period=2;
	}else
	if(period>65536){
		period=65536;
	}
	period-=1;

	if(duty>65535){
		duty=65535;
	}

	switch(ch){
	case	0:
		SetDCOC1PWM(duty,period);
		break;
	case	1:
		SetDCOC2PWM(duty,period);
		break;
	case	2:
		SetDCOC3PWM(duty,period);
		break;
	case	3:
		SetDCOC4PWM(duty,period);
		break;
	case	4:
		SetDCOC5PWM(duty,period);
		break;
	default:
		break;
	}
	return;
}

static void set_pwm_config(uint8_t config)
{
	static uint8_t	last_config=0;
//	uint16_t		width,duty,i;

	if(!last_config && config){
		//initialize pwm
		OpenTimer2(T2_ON|PWM_PRESCALER,PWM_TIMER_PERIOD-1);
	}else
	if(last_config && !config){
		//terminate pwm
		CloseOC1();
		CloseOC2();
		CloseOC3();
		CloseOC4();
		CloseOC5();
		CloseTimer2();
	}

	if(!(last_config&0x01) && config&0x01){
		OpenOC1(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(0);
	}else
	if(last_config&0x01 && !(config&0x01)){
		SetDCOC1PWM(0,0);
		CloseOC1();
	}

	if(!(last_config&0x02) && config&0x02){
		OpenOC2(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(1);
	}else
	if(last_config&0x02 && !(config&0x02)){
		SetDCOC2PWM(0,0);
		CloseOC2();
	}

	if(!(last_config&0x04) && config&0x04){
		OpenOC3(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(2);
	}else
	if(last_config&0x04 && !(config&0x04)){
		SetDCOC3PWM(0,0);
		CloseOC3();
	}

	if(!(last_config&0x08) && config&0x08){
		OpenOC4(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(3);
	}else
	if(last_config&0x08 && !(config&0x08)){
		SetDCOC4PWM(0,0);
		CloseOC4();
	}

	if(!(last_config&0x10) && config&0x10){
		OpenOC5(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(4);
	}else
	if(last_config&0x10 && !(config&0x10)){
		SetDCOC5PWM(0,0);
		CloseOC5();
	}

	pwm_setup(config);
	last_config=config;

	return;
}

static void initconfig(void)
{
	pio_input_config=
	uart_rx_config=
	lowbat_notify_config=
		0;

	return;
}

static uint8_t i2c_start_next_rw=0;

static int ble_i2c_startstop(uint8_t startstop)
{
	switch(startstop){
	case	0:
		//i2c stop
		i2c_stop();
		break;
	case	1:
#if	1
		i2c_start_next_rw=1;
#else
		//i2c start
		if(i2c_waitbusfree()){
			//^CAEg
			return -1;
		}
	
		if(i2c_start()){
			return -1;
		}
#endif
		break;
	case	2:
		//i2c repeated start
		if(i2c_start()){
			return -1;
		}
		break;
	}
	return 0;
}

static void ble_i2c_write(uint8_t *buffer,uint16_t buffer_size)
{
	uint8_t	len;

	len=*buffer++;

	if(i2c_start_next_rw){
		i2c_start_next_rw=0;
		if(i2c_waitbusfree()){
			//^CAEg
			return;
		}
	//	*buffer&=0xfe;
		if(send_ctrl(*buffer++)){
			//
			return;
		}
		len--;
	}

	for(;len;len--){
		i2c_write(*buffer++);
	}
//	i2c_stop();
}

static void ble_i2c_read(uint8_t *buffer,uint16_t buffer_size)
{
	uint8_t	len;
	uint8_t *p;

	len=buffer[0];
	if(i2c_start_next_rw){
		i2c_start_next_rw=0;
		if(i2c_waitbusfree()){
			//^CAEg
			return;
		}
		if(send_ctrl(buffer[1])){
			//
			return;
		}
	}else{
		i2c_write(buffer[1]);
	}


	for(p=i2c_recvbuf;len;len--){
		*p=i2c_read();
		if(len>1){
			i2c_ack();
		}
		p++;
	}
	i2c_nak() ;

	i2c_recvlen=buffer[0];
	return;
}

// read requests
static uint16_t att_read_callback(uint16_t handle, uint16_t offset, uint8_t * buffer, uint16_t buffer_size){
	uint16_t	ret=0,i;
	int			c;

	if(buffer){
		log_info("READ Callback, handle %04x\n", handle);
		led1_on_mac();
	}
	switch(handle){
	case	HANDLE_PIO_SETTING_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=pio_setting;
		}
		break;
	case	HANDLE_PIO_PULLUP_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=pio_pullup;
		}
		break;
	case	HANDLE_PIO_INPUT_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=pio_input;
		}
		break;
	case	HANDLE_AI_READ0_VALUE:
		ret=2;
		if(buffer && buffer_size>=2){
			i=advalue2mv(advalue[0]);
			buffer[0]=i>>8;
			buffer[1]=i&0xff;
		}
		break;
	case	HANDLE_AI_READ1_VALUE:
		ret=2;
		if(buffer && buffer_size>=2){
			i=advalue2mv(advalue[1]);
			buffer[0]=i>>8;
			buffer[1]=i&0xff;
		}
		break;
	case	HANDLE_AI_READ2_VALUE:
		break;
	case	HANDLE_I2C_READ_VALUE:
		//received bytes
		ret=i2c_recvlen;
		if(buffer){
			if(buffer_size>=i2c_recvlen){
				for(i=0;i<i2c_recvlen;i++){
					buffer[i]=i2c_recvbuf[i];
				}
				for(;i<buffer_size;i++){
					buffer[i]=0;	//padding
				}
			}
			i2c_recvlen=0;
		}
		break;
	case	HANDLE_UART_CONFIG_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=uart_config;
		}
		break;
	case	HANDLE_UART_BAUDRATE_VALUE:
		ret=2;
		if(buffer && buffer_size){
			buffer[0]=uart_baudrate>>8;
			buffer[1]=uart_baudrate&0xff;
		}
		break;
	case	HANDLE_UART_TX_VALUE:
		ret=uart_lasttx_size;
		if(buffer && buffer_size){
			for(i=0;i<buffer_size;i++){
				if(i<uart_lasttx_size){
					buffer[i]=uart_lasttx_str[i];
				}else{
					buffer[i]=0;	//padding
				}
			}
		}
		break;
	case	HANDLE_UART_RX_VALUE:
		ret=UART1GetRXCnt();
#if	UART_PACKET_MULTIPLE_BYTES
		if(ret>=(att_connection.mtu-3)){
			ret=(att_connection.mtu-3);
		}
#else
		if(ret){
			ret=1;
		}
#endif
		if(buffer && buffer_size){
            log_info("UART1 RX: ");
			for(i=0;i<buffer_size;i++){
				if(i<ret){
					c=UART1GetChar();
					if(c<0){
						c=0;			//padding
					}
				}else{
					c=0;				//padding
				}
				buffer[i]=(uint8_t)c;
			}
		}
		break;
	case	HANDLE_LOWBAT_NOTIFY_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=0;		//no lowbat
		}
		break;
	case	HANDLE_PIO_INPUT_CONFIG:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=pio_input_config;
		}
		break;
	case	HANDLE_UART_RX_CONFIG:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=uart_rx_config;
		}
		break;
	case	HANDLE_LOWBAT_NOTIFY_CONFIG:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=lowbat_notify_config;
		}
		break;
	case	HANDLE_MACHINE_ID_VALUE:
		ret=4;
		if(buffer && buffer_size>=4){
			net_store_32(buffer,0,MACHINE_ID);
		}
		break;
	case	HANDLE_UART_POLARITY_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=uart1_polarity;
		}
		break;
	default:
		break;
    }
	return ret;
}

// write requests
static void att_write_callback(uint16_t handle, uint16_t transaction_mode, uint16_t offset, uint8_t *buffer, uint16_t buffer_size, signature_t * signature){
	uint16_t	i;
	uint32_t	l;

    log_info("WRITE Callback, handle %04x\n", handle);
	led1_on_mac();
    switch(handle){
	case	HANDLE_PIO_SETTING_VALUE:
		if(buffer_size){
			pio_setting=buffer[0];
			gpio_setup(pio_setting);
		}
		break;
	case	HANDLE_PIO_PULLUP_VALUE:
		if(buffer_size){
			pio_pullup=buffer[0];
			gpio_setpullup(pio_pullup);
		}
		break;
	case	HANDLE_PIO_OUTPUT_VALUE:
		if(buffer_size){
			gpio_output(buffer[0]);
		}
		break;
	case	HANDLE_PWM_CONFIG_VALUE:
		if(buffer_size){
			pwm_config=buffer[0];
			set_pwm_config(pwm_config);
		}
		break;
	case	HANDLE_PWM_PARAM_VALUE:
		if(buffer_size>=5){
			i=buffer[0];
			l =((uint32_t)buffer[1])<<24;
			l|=((uint32_t)buffer[2])<<16;
			l|=((uint32_t)buffer[3])<< 8;
			l|=((uint32_t)buffer[4])<< 0;
			if(i<8){
				pwm_period[i]=l;
				set_pwm(i);
			}
		}
		break;
	case	HANDLE_PWM_DUTY_VALUE:
		if(buffer_size>=5){
			i=buffer[0];
			l =((uint32_t)buffer[1])<<24;
			l|=((uint32_t)buffer[2])<<16;
			l|=((uint32_t)buffer[3])<< 8;
			l|=((uint32_t)buffer[4])<< 0;
			if(i<8){
				pwm_duty[i]=l;
				set_pwm(i);
			}
		}
		break;
	case	HANDLE_I2C_CONFIG_VALUE:
		//write 0:OFF 1:100kHz 2:400kHz
		if(buffer_size){
			i2c_config=buffer[0];
			if(buffer[0]){
				open_i2c(buffer[0]);
			}
		}
		break;
	case	HANDLE_I2C_START_STOP_VALUE:
		//write 0:stop 1:start condition 2:repeated start condition
		if(buffer_size){
			ble_i2c_startstop(buffer[0]);
		}
		break;
	case	HANDLE_I2C_WRITE_VALUE:	//write command
		//write +0:write length(from next byte) +1:(addr<<1)|(R:1/W:0) +2`:Write bytes
		if(buffer_size>=3){
			ble_i2c_write(buffer,buffer_size);
		}
		break;
	case	HANDLE_I2C_READ_PARAM_VALUE:	//read command
		//write +0:read length(from next byte) +1:(addr<<1)|(R:1/W:0)
		if(buffer_size>=2){
			ble_i2c_read(buffer,buffer_size);
		}
		break;
	case	HANDLE_UART_CONFIG_VALUE:
		if(buffer_size){
			uart_config=buffer[0];
			if(uart_config){
				//UART̃|[ggpioŎgp֎~ɂ
				portmap[UART_MAPNO1].enable=0;
				portmap[UART_MAPNO2].enable=0;
#if	UART_FIX_BAUDRATE
				UART1Init(UART_DEFAULT_BAUDRATE);
#else
				UART1Init(uart_baudrate*240LU);
#endif
			}else{
				UART1Term();
				//UART̃|[ggpioŎgp\ɂ
				portmap[UART_MAPNO1].enable=1;
				portmap[UART_MAPNO2].enable=1;
			}
		}
		break;
	case	HANDLE_UART_BAUDRATE_VALUE:
		if(buffer_size>=2){
			uart_baudrate=((uint16_t)(buffer[0]))<<8;
			uart_baudrate+=buffer[1];
		}
		break;
	case	HANDLE_UART_TX_VALUE:
		if(buffer_size){
            log_info("UART1 TX: ");
			for(i=0;i<buffer_size;i++){
				log_info("%02x ",buffer[i]);
				UART1PutChar(buffer[i]);
			}
			log_info("\n");
		}
		break;
	case	HANDLE_PIO_INPUT_CONFIG:
		if(buffer_size){
			pio_input_config=buffer[0];
		}
		break;
	case	HANDLE_UART_RX_CONFIG:
		if(buffer_size){
			uart_rx_config=buffer[0];
		}
		break;
	case	HANDLE_HARDWARE_RESET_VALUE:
		if(buffer_size){
			if(buffer[0]){
				Reset();
			}
		}
		break;
	case	HANDLE_LOWBAT_NOTIFY_CONFIG:
		if(buffer_size){
			lowbat_notify_config=buffer[0];
		}
		break;
	case	HANDLE_UART_POLARITY_VALUE:
		if(buffer_size){
			uart1_polarity=buffer[0];
		}
		break;
	default:
		break;
    }
}

// check notification
void check_notification(void)
{
	int i,j,c;
	uint8_t	notify_buf[20];
	static uint8_t	pio_last_input=0;

	if(connection_status){
		if(
			uart_rx_config&FLAG_NOTIFICATION
		&&	UART1GetRXCnt()
		&&	hci_can_send_packet_now(HCI_ACL_DATA_PACKET)		//M]T͂邩
		){
#if	UART_PACKET_MULTIPLE_BYTES
			j=att_connection.mtu-3;
#else
			j=1;
#endif
			for(i=0;i<j;i++){
				c=UART1GetChar();
				if(c<0){
					break;
				}
				notify_buf[i]=(uint8_t)c;
			}
			if(i){
				log_info("UART RX NOTIFICATION\n");
				att_response_size=att_prepare_handle_value_notification(&att_connection,HANDLE_UART_RX_VALUE,notify_buf,i,att_response_buffer);
				att_try_respond();
			}
		}

		pio_input=gpio_input();
		if(pio_input_config&FLAG_NOTIFICATION){
			if(
				pio_input!=pio_last_input
			&&	hci_can_send_packet_now(HCI_ACL_DATA_PACKET)		//M]T͂邩
			){
				log_info("PIO INPUT NOTIFICATION\n");
				notify_buf[0]=pio_input;
				att_response_size=att_prepare_handle_value_notification(&att_connection,HANDLE_PIO_INPUT_VALUE,notify_buf,1,att_response_buffer);
				att_try_respond();
				pio_last_input=pio_input;
			}
		}else{
			pio_last_input=pio_input;
		}

		if(lowbat_notify_config&FLAG_NOTIFICATION){
		}
	}
}

// main
int blemain(void)
{
	/// GET STARTED with BTstack ///
	btstack_memory_init();
    run_loop_init(RUN_LOOP_EMBEDDED);
	
    // init HCI
	hci_transport_t    * transport = hci_transport_picusb_instance();
	bt_control_t       * control   = NULL;
    hci_uart_config_t  * config    = NULL;
    remote_device_db_t * remote_db = (remote_device_db_t *) &remote_device_db_memory;
	hci_init(transport, config, control, remote_db);
	
    // use eHCILL
    // bt_control_cc256x_enable_ehcill(1);
    
    // set up l2cap_le
    l2cap_init();
    l2cap_register_fixed_channel(att_packet_handler, L2CAP_CID_ATTRIBUTE_PROTOCOL);
    l2cap_register_packet_handler(packet_handler);
    
    // set up ATT
	log_info("profile_data %u bytes.\n",sizeof(profile_data));
    att_set_db(profile_data);
    att_set_write_callback(att_write_callback);
    att_set_read_callback(att_read_callback);
    att_dump_attributes();
    att_connection.mtu = 27;
    
	log_info("Run...\n\r");

    // turn on!
	hci_power_control(HCI_POWER_ON);

    // go!
    run_loop_execute();	
    
    // happy compiler!
    return 0;
}

